/**
 * Copyright (C) 2007  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.novabpm.identity.auth;

import java.security.Principal;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.StringTokenizer;

import javax.management.remote.JMXPrincipal;
import javax.security.auth.Destroyable;
import javax.security.auth.Subject;
import javax.security.auth.callback.Callback;
import javax.security.auth.callback.CallbackHandler;
import javax.security.auth.callback.NameCallback;
import javax.security.auth.callback.PasswordCallback;
import javax.security.auth.login.FailedLoginException;
import javax.security.auth.login.LoginException;
import javax.security.auth.spi.LoginModule;

/**
 * @author "Pierre Vigneras"
 * @date Jan 11, 2008
 */
public class PlainLoginModule implements LoginModule {
	private static final String NAME_PROMPT = "Name: ";
	private static final String PASSWORD_PROMPT = "Password: ";

	private static final String JAVAX_SECURITY_AUTH_LOGIN_PASSWORD = "javax.security.auth.login.password";
	private static final String JAVAX_SECURITY_AUTH_LOGIN_NAME = "javax.security.auth.login.name";

	/**
	 * Property key for the debug flag. Defined to be "debug".
	 * 
	 * Property Value. If set, should be either "true" or "false". Default is
	 * "false".
	 */
	public static final String DEBUG_OPTION_NAME = "debug";
	private static final Object LOGINS_OPTION_NAME = "logins";
	private static final Object PASSWORDS_OPTION_NAME = "passwords";
	private Subject subject = null;
	private CallbackHandler callbackHandler = null;
	private Map<String, Object> sharedState;
	private Map<String, Object> options = null;
	private boolean debug = false;

	private String id;
	private Map<String, String> logins;

	/**
	 * Initialize this LoginModule. This method is called by the LoginContext
	 * after this LoginModule has been instantiated. The purpose of this method
	 * is to initialize this LoginModule with the relevant information. If this
	 * LoginModule does not understand any of the data stored in sharedState or
	 * options parameters, they can be ignored.
	 * 
	 * @param subject
	 *            the Subject to be authenticated.
	 * @param callbackHandler
	 *            a CallbackHandler for communicating with the end user
	 *            (prompting for usernames and passwords, for example).
	 * @param sharedState
	 *            state shared with other configured LoginModules.
	 * @param options
	 *            options specified in the login Configuration for this
	 *            particular LoginModule.
	 */
	public void initialize(Subject subject, CallbackHandler callbackHandler,
			Map<String, ?> sharedState, Map<String, ?> options) {

		this.subject = subject;
		this.callbackHandler = callbackHandler;
		this.sharedState = (Map<String, Object>) sharedState;
		this.options = (Map<String, Object>) options;
		final String debugFlag = (String) options.get(DEBUG_OPTION_NAME);
		if (debugFlag != null) {
			debug = Boolean.valueOf(debugFlag);
		}
	}

	/**
	 * Method to authenticate a Subject (phase 1). The implementation of this
	 * method authenticates a Subject. For example, it may prompt for Subject
	 * information such as a username and password and then attempt to verify
	 * the password. This method saves the result of the authentication attempt
	 * as private state within the LoginModule.
	 * 
	 * @return true if the authentication succeeded, or false if this
	 *         LoginModule should be ignored.
	 * @throws LoginException
	 *             if the authentication fails
	 */
	public boolean login() throws LoginException {
		if (debug)
			System.err.println("[" + PlainLoginModule.class.getName()
					+ "] login() - preparing - step 1");
		if (logins == null) {
			final String loginList = (String) options.get(LOGINS_OPTION_NAME);
			final String passwordList = (String) options
					.get(PASSWORDS_OPTION_NAME);
			if (loginList == null)
				throw new MissingOptionException("Option: "
						+ LOGINS_OPTION_NAME
						+ " should be defined as a comma separated login list");
			if (passwordList == null)
				throw new MissingOptionException(
						"Option: "
								+ PASSWORDS_OPTION_NAME
								+ " should be defined as a comma separated password list");
			final StringTokenizer loginTokenizer = new StringTokenizer(
					loginList, ",");
			final StringTokenizer passwordTokenizer = new StringTokenizer(
					passwordList, ",");
			if (loginTokenizer.countTokens() != passwordTokenizer.countTokens())
				throw new BadLoginOptionException(
						"login list and password list should have the same number of elements!");
			logins = new HashMap<String, String>();
			while (loginTokenizer.hasMoreTokens()) {
				logins.put(loginTokenizer.nextToken().trim(), passwordTokenizer
						.nextToken().trim());
			}
		}
		try {
			String name = (String) sharedState
					.get(JAVAX_SECURITY_AUTH_LOGIN_NAME);
			String password = (String) sharedState
					.get(JAVAX_SECURITY_AUTH_LOGIN_PASSWORD);
			final List<Callback> callbacks = new ArrayList<Callback>();
			final NameCallback nameCallback = new NameCallback(NAME_PROMPT);
			final PasswordCallback passwordCallback = new PasswordCallback(
					PASSWORD_PROMPT, false);
			if (name == null)
				callbacks.add(nameCallback);
			if (password == null)
				callbacks.add(passwordCallback);
			if (!callbacks.isEmpty()) {
				if (debug)
					System.err.println("[" + PlainLoginModule.class.getName()
							+ "] login() - callback - step 2");
				callbackHandler.handle(callbacks.toArray(new Callback[0]));
				if (name == null)
					name = nameCallback.getName();
				if (password == null) {
					password = new String(passwordCallback.getPassword());
					passwordCallback.clearPassword();
				}
			}
			if (debug) {
				System.err.println("[" + PlainLoginModule.class.getName()
						+ "] login() - authenticating - step 3");
			}

			if (logins.containsKey(name) && logins.containsValue(password)) {
				this.id = name;
			}
			// TODO: what is Step 4 then?!
			if (debug)
				System.err.println("[" + PlainLoginModule.class.getName()
						+ "] login() - storing data - step 5");
			sharedState.put(JAVAX_SECURITY_AUTH_LOGIN_NAME, name);
			sharedState.put(JAVAX_SECURITY_AUTH_LOGIN_PASSWORD, password);
			if (debug)
				System.err.println("[" + PlainLoginModule.class.getName()
						+ "] login() - returning - step 6");
			if (id == null)
				throw new FailedLoginException(
						"User unknown or wrong password!");
			return true;
		} catch (Exception e) {
			final LoginException le = new LoginException();
			le.initCause(e);
			throw le;
		}
	}

	/**
	 * Method to commit the authentication process (phase 2). This method is
	 * called if the LoginContext's overall authentication succeeded (the
	 * relevant REQUIRED, REQUISITE, SUFFICIENT and OPTIONAL LoginModules
	 * succeeded). If this LoginModule's own authentication attempt succeeded
	 * (checked by retrieving the private state saved by the login method), then
	 * this method associates relevant Principals and Credentials with the
	 * Subject located in the LoginModule. If this LoginModule's own
	 * authentication attempted failed, then this method removes/destroys any
	 * state that was originally saved.
	 * 
	 * @return true if this method succeeded, or false if this LoginModule
	 *         should be ignored.
	 * @throws LoginException
	 *             if the commit fails
	 */
	public boolean commit() throws LoginException {
		if (id == null)
			throw new FailedLoginException("User unknown or wrong password!");
		final Set<Principal> principals = subject.getPrincipals();
		// We use JMXPrincipal as it is the single Principal implementation available 
    // in the whole J2SE 5.
		principals.add(new JMXPrincipal(id));
		return true;
	}

	/**
	 * Method to abort the authentication process (phase 2). This method is
	 * called if the LoginContext's overall authentication failed. (the relevant
	 * REQUIRED, REQUISITE, SUFFICIENT and OPTIONAL LoginModules did not
	 * succeed). If this LoginModule's own authentication attempt succeeded
	 * (checked by retrieving the private state saved by the login method), then
	 * this method cleans up any state that was originally saved.
	 * 
	 * @return true if this method succeeded, or false if this LoginModule
	 *         should be ignored.
	 * @throws LoginException
	 *             if the abort fails
	 */
	public boolean abort() throws LoginException {
		if (debug)
			System.err.println("[" + PlainLoginModule.class.getName()
					+ "] abort()");
		if (id == null)
			return false;
		subject = null;
		id = null;
		return true;
	}

	/**
	 * Method which logs out a Subject. An implementation of this method might
	 * remove/destroy a Subject's Principals and Credentials.
	 * 
	 * @return true if this method succeeded, or false if this LoginModule
	 *         should be ignored.
	 * @throws LoginException
	 *             if the logout fails
	 */
	public boolean logout() throws LoginException {
		if (id != null) {
			if (debug)
				System.err.println("[" + PlainLoginModule.class.getName()
						+ "] logout() - removing principals");
			// Remove only principals added by our commit method
			final Set<Principal> principals = subject.getPrincipals();
			for (Principal p : principals) {
				if (p instanceof BonitaPrincipal) {
					if (debug)
						System.err.println("["
								+ PlainLoginModule.class.getName()
								+ "] logout() - removing principal: " + p);
					principals.remove(p);
				}
			}
			if (debug)
				System.err.println("[" + PlainLoginModule.class.getName()
						+ "] logout() - destroying/removing credentials");
			// Remove/destroy only credentials added by our commit method
			final Set<Object> credentials = subject.getPublicCredentials();
			for (Object o : credentials) {
				if (o instanceof Destroyable) {
					if (debug)
						System.err.println("["
								+ PlainLoginModule.class.getName()
								+ "] logout() - destroying credential: " + o);
					// Bug: only from this module !!
					// ((Destroyable) o).destroy();
				}
				if (!subject.isReadOnly()) {
					if (debug)
						System.err.println("["
								+ PlainLoginModule.class.getName()
								+ "] logout() - removing credential: " + o);
					credentials.remove(o);
				}
			}
		}
		return true;
	}
}
